@@ -19,10 +19,11 @@ from pathlib import PurePath
 from collections import OrderedDict
 from .mesonlib import MesonException
 from .mesonlib import default_libdir, default_libexecdir, default_prefix
+from .wrap import WrapMode
 import ast
 import argparse
 
-version = '0.47.2'
+version = '0.48.1'
 backendlist = ['ninja', 'vs', 'vs2010', 'vs2015', 'vs2017', 'xcode']
 
 default_yielding = False
@@ -212,6 +213,7 @@ class CoreData:
         }
         self.test_guid = str(uuid.uuid4()).upper()
         self.regen_guid = str(uuid.uuid4()).upper()
+        self.install_guid = str(uuid.uuid4()).upper()
         self.target_guids = {}
         self.version = version
         self.init_builtins()
@@ -221,7 +223,7 @@ class CoreData:
         self.base_options = {}
         self.external_preprocess_args = {} # CPPFLAGS only
         self.cross_file = self.__load_cross_file(options.cross_file)
-        self.wrap_mode = options.wrap_mode
+        self.wrap_mode = options.wrap_mode if options.wrap_mode is not None else WrapMode.default
         self.compilers = OrderedDict()
         self.cross_compilers = OrderedDict()
         self.deps = OrderedDict()
@@ -334,7 +336,6 @@ class CoreData:
                     'Default project to execute in Visual Studio',
                     '')
 
-
     def get_builtin_option(self, optname):
         if optname in self.builtins:
             return self.builtins[optname].value
@@ -350,6 +351,51 @@ class CoreData:
             raise RuntimeError('Tried to set unknown builtin option %s.' % optname)
         self.builtins[optname].set_value(value)
 
+        # Make sure that buildtype matches other settings.
+        if optname == 'buildtype':
+            self.set_others_from_buildtype(value)
+        else:
+            self.set_buildtype_from_others()
+
+    def set_others_from_buildtype(self, value):
+        if value == 'plain':
+            opt = '0'
+            debug = False
+        elif value == 'debug':
+            opt = '0'
+            debug = True
+        elif value == 'debugoptimized':
+            opt = '2'
+            debug = True
+        elif value == 'release':
+            opt = '3'
+            debug = False
+        elif value == 'minsize':
+            opt = 's'
+            debug = True
+        else:
+            assert(value == 'custom')
+            return
+        self.builtins['optimization'].set_value(opt)
+        self.builtins['debug'].set_value(debug)
+
+    def set_buildtype_from_others(self):
+        opt = self.builtins['optimization'].value
+        debug = self.builtins['debug'].value
+        if opt == '0' and not debug:
+            mode = 'plain'
+        elif opt == '0' and debug:
+            mode = 'debug'
+        elif opt == '2' and debug:
+            mode = 'debugoptimized'
+        elif opt == '3' and not debug:
+            mode = 'release'
+        elif opt == 's' and debug:
+            mode = 'minsize'
+        else:
+            mode = 'custom'
+        self.builtins['buildtype'].set_value(mode)
+
     def validate_option_value(self, option_name, override_value):
         for opts in (self.builtins, self.base_options, self.compiler_options, self.user_options):
             if option_name in opts:
@@ -390,8 +436,7 @@ class CoreData:
             if k == 'prefix':
                 pass
             elif k in self.builtins:
-                tgt = self.builtins[k]
-                tgt.set_value(self.sanitize_dir_option_value(prefix, k, v))
+                self.set_builtin_option(k, v)
             elif k in self.backend_options:
                 tgt = self.backend_options[k]
                 tgt.set_value(v)
@@ -545,31 +590,33 @@ def parse_cmd_line_options(args):
             delattr(args, name)
 
 builtin_options = {
-    'buildtype':  [UserComboOption, 'Build type to use.', ['plain', 'debug', 'debugoptimized', 'release', 'minsize'], 'debug'],
-    'strip':      [UserBooleanOption, 'Strip targets on install.', False],
-    'unity':      [UserComboOption, 'Unity build.', ['on', 'off', 'subprojects'], 'off'],
-    'prefix':     [UserStringOption, 'Installation prefix.', default_prefix()],
-    'libdir':     [UserStringOption, 'Library directory.', default_libdir()],
-    'libexecdir': [UserStringOption, 'Library executable directory.', default_libexecdir()],
-    'bindir':     [UserStringOption, 'Executable directory.', 'bin'],
-    'sbindir':    [UserStringOption, 'System executable directory.', 'sbin'],
-    'includedir': [UserStringOption, 'Header file directory.', 'include'],
-    'datadir':    [UserStringOption, 'Data file directory.', 'share'],
-    'mandir':     [UserStringOption, 'Manual page directory.', 'share/man'],
-    'infodir':    [UserStringOption, 'Info page directory.', 'share/info'],
-    'localedir':  [UserStringOption, 'Locale data directory.', 'share/locale'],
-    'sysconfdir':      [UserStringOption, 'Sysconf data directory.', 'etc'],
-    'localstatedir':   [UserStringOption, 'Localstate data directory.', 'var'],
-    'sharedstatedir':  [UserStringOption, 'Architecture-independent data directory.', 'com'],
-    'werror':          [UserBooleanOption, 'Treat warnings as errors.', False],
-    'warning_level':   [UserComboOption, 'Compiler warning level to use.', ['1', '2', '3'], '1'],
-    'layout':          [UserComboOption, 'Build directory layout.', ['mirror', 'flat'], 'mirror'],
-    'default_library': [UserComboOption, 'Default library type.', ['shared', 'static', 'both'], 'shared'],
-    'backend':         [UserComboOption, 'Backend to use.', backendlist, 'ninja'],
-    'stdsplit':        [UserBooleanOption, 'Split stdout and stderr in test logs.', True],
-    'errorlogs':       [UserBooleanOption, "Whether to print the logs from failing tests.", True],
-    'install_umask':   [UserUmaskOption, 'Default umask to apply on permissions of installed files.', '022'],
-    'auto_features':   [UserFeatureOption, "Override value of all 'auto' features.", 'auto'],
+    'buildtype':  [UserComboOption, 'Build type to use', ['plain', 'debug', 'debugoptimized', 'release', 'minsize', 'custom'], 'debug'],
+    'strip':      [UserBooleanOption, 'Strip targets on install', False],
+    'unity':      [UserComboOption, 'Unity build', ['on', 'off', 'subprojects'], 'off'],
+    'prefix':     [UserStringOption, 'Installation prefix', default_prefix()],
+    'libdir':     [UserStringOption, 'Library directory', default_libdir()],
+    'libexecdir': [UserStringOption, 'Library executable directory', default_libexecdir()],
+    'bindir':     [UserStringOption, 'Executable directory', 'bin'],
+    'sbindir':    [UserStringOption, 'System executable directory', 'sbin'],
+    'includedir': [UserStringOption, 'Header file directory', 'include'],
+    'datadir':    [UserStringOption, 'Data file directory', 'share'],
+    'mandir':     [UserStringOption, 'Manual page directory', 'share/man'],
+    'infodir':    [UserStringOption, 'Info page directory', 'share/info'],
+    'localedir':  [UserStringOption, 'Locale data directory', 'share/locale'],
+    'sysconfdir':      [UserStringOption, 'Sysconf data directory', 'etc'],
+    'localstatedir':   [UserStringOption, 'Localstate data directory', 'var'],
+    'sharedstatedir':  [UserStringOption, 'Architecture-independent data directory', 'com'],
+    'werror':          [UserBooleanOption, 'Treat warnings as errors', False],
+    'warning_level':   [UserComboOption, 'Compiler warning level to use', ['1', '2', '3'], '1'],
+    'layout':          [UserComboOption, 'Build directory layout', ['mirror', 'flat'], 'mirror'],
+    'default_library': [UserComboOption, 'Default library type', ['shared', 'static', 'both'], 'shared'],
+    'backend':         [UserComboOption, 'Backend to use', backendlist, 'ninja'],
+    'stdsplit':        [UserBooleanOption, 'Split stdout and stderr in test logs', True],
+    'errorlogs':       [UserBooleanOption, "Whether to print the logs from failing tests", True],
+    'install_umask':   [UserUmaskOption, 'Default umask to apply on permissions of installed files', '022'],
+    'auto_features':   [UserFeatureOption, "Override value of all 'auto' features", 'auto'],
+    'optimization':    [UserComboOption, 'Optimization level', ['0', 'g', '1', '2', '3', 's'], '0'],
+    'debug':           [UserBooleanOption, 'Debug', True]
 }
 
 # Special prefix-dependent defaults for installation directories that reside in
