@@ -158,6 +158,11 @@ def platform_fix_name(fname, compiler, env):
         if compiler == 'cl':
             return None
 
+    if fname.startswith('?cygwin:'):
+        fname = fname[8:]
+        if compiler == 'cl' or not mesonlib.for_cygwin(env.is_cross_build(), env):
+            return None
+
     return fname
 
 def validate_install(srcdir, installdir, compiler, env):
@@ -183,17 +188,11 @@ def validate_install(srcdir, installdir, compiler, env):
             expected[fname] = True
     for (fname, found) in expected.items():
         if not found:
-            # Ignore missing PDB files if we aren't using cl
-            if fname.endswith('.pdb') and compiler != 'cl':
-                continue
             ret_msg += 'Expected file {0} missing.\n'.format(fname)
     # Check if there are any unexpected files
     found = get_relative_files_list_from_dir(installdir)
     for fname in found:
-        # Windows-specific tests check for the existence of installed PDB
-        # files, but common tests do not, for obvious reasons. Ignore any
-        # extra PDB files found.
-        if fname not in expected and not fname.endswith('.pdb') and compiler == 'cl':
+        if fname not in expected:
             ret_msg += 'Extra file {0} found.\n'.format(fname)
     return ret_msg
 
@@ -294,16 +293,16 @@ def run_test(skipped, testdir, extra_args, compiler, backend, flags, commands, s
                 mlog.shutdown() # Close the log file because otherwise Windows wets itself.
 
 def pass_prefix_to_test(dirname):
-    if '40 prefix' in dirname:
+    if '39 prefix absolute' in dirname:
         return False
     return True
 
 def pass_libdir_to_test(dirname):
     if '8 install' in dirname:
         return False
-    if '39 libdir' in dirname:
+    if '38 libdir must be inside prefix' in dirname:
         return False
-    if '201 install_mode' in dirname:
+    if '196 install_mode' in dirname:
         return False
     return True
 
@@ -636,10 +635,34 @@ def check_file(fname):
         linenum += 1
 
 def check_format():
+    check_suffixes = {'.c',
+                      '.cpp',
+                      '.cxx',
+                      '.cc',
+                      '.rs',
+                      '.f90',
+                      '.vala',
+                      '.d',
+                      '.s',
+                      '.m',
+                      '.mm',
+                      '.asm',
+                      '.java',
+                      '.txt',
+                      '.py',
+                      '.swift',
+                      '.build',
+                      '.md',
+                      }
     for (root, _, files) in os.walk('.'):
-        for file in files:
-            if file.endswith('.py') or file.endswith('.build') or file == 'meson_options.txt':
-                fullname = os.path.join(root, file)
+        if '.dub' in root: # external deps are here
+            continue
+        for fname in files:
+            if os.path.splitext(fname)[1].lower() in check_suffixes:
+                bn = os.path.basename(fname)
+                if bn == 'sitemap.txt' or bn == 'meson-test-run.txt':
+                    continue
+                fullname = os.path.join(root, fname)
                 check_file(fullname)
 
 def check_meson_commands_work():
